require 'eventmachine'
require 'amqp'
require 'rflow'
require 'time'

class RFlow
  # @!parse
  #   # Fake classes in this tree to document the actual message types.
  #   class Message
  #     # Fake classes in this tree to document the actual message types.
  #     class Data
  #       # AMQP messages.
  #       module AMQP
  #         # RFlow format defined for AMQP messages which can be emitted by
  #         # {RFlow::Components::AMQP::Subscriber}. Of course the real class
  #         # is {RFlow::Message} with type +RFlow::Message::Data::AMQP::Message+.
  #         class Message
  #           # @!attribute header
  #           #   The header of the message.
  #           #   @return [String]
  #
  #           # @!attribute payload
  #           #   The payload of the message.
  #           #   @return [String]
  #
  #           # Just here to force Yard to create documentation.
  #           # @!visibility private
  #           def initialize; end
  #         end
  #       end
  #     end
  #   end

  # Components.
  module Components
    module AMQP
      # Component that subscribes to an AMQP topic and issues +RFlow::Message+s
      # of type {RFlow::Message::Data::AMQP::Message} and +RFlow::Message::Data::Raw+
      # when it receives messages on the AMQP topic.
      #
      # By default will use an exclusive, non-durable, auto-deleting,
      # randomly named queue for subscribing to topic messages with an
      # empty-string pattern.
      #
      # Accepts config parameters:
      # - +server+ - server address
      # - +port+ - server port
      # - +username+ - server username
      # - +password+ - server password
      # - +vhost+ - vhost
      # - +reconnect_interval+ - how long to wait when disconnected before reconnecting
      # - +queue_name+ - AMQP topic name to monitor
      # - +queue_passive+ - +queue_passive+ parameter to AMQP
      # - +queue_durable+ - +queue_durable+ parameter to AMQP
      # - +queue_exclusive+ - +queue_exclusive+ parameter to AMQP
      # - +queue_auto_delete+ - +queue_auto_delete+ parameter to AMQP
      # - +binding_pattern+ - AMQP routing key
      class Subscriber < RFlow::Component
        # @!attribute [r] amqp_port
        #   Outputs +RFlow::Message+s of type {RFlow::Message::Data::AMQP::Message}
        #   when receiving an AMQP message.
        #
        #   @return [RFlow::Component::OutputPort]
        output_port :amqp_port

        # @!attribute [r] raw_port
        #   Outputs +RFlow::Message+s of type +RFlow::Message::Data::Raw+
        #   when receiving an AMQP message. Messages contain just the bytes.
        #
        #   @return [RFlow::Component::OutputPort]
        output_port :raw_port

        # @!attribute [r] log_port
        #   Outputs +RFlow::Message+s of type {RFlow::Message:Data::Log} if there
        #   are problems with the AMQP subscriber..
        #   @return [RFlow::Component::OutputPort]
        output_port :log_port

        # @!visibility private
        attr_accessor :config, :queue_config

        # Default config.
        DEFAULT_CONFIG = {
          'server' => '127.0.0.1',
          'port' => 5672,
          'username' => 'guest',
          'password' => 'guest',
          'vhost' => '/',
          'reconnect_interval' => 20,

          'queue_name' => 'asdf',
          'queue_passive' => false,
          'queue_durable' => false,
          'queue_exclusive' => true,
          'queue_auto_delete' => true,

          'binding_pattern' => '',
        }

        # RFlow-called method at startup.
        # @return [void]
        def configure!(config)
          @config = DEFAULT_CONFIG.merge config
          @config['port'] = @config['port'].to_i
          @config['reconnect_interval'] = @config['reconnect_interval'].to_i

          ['durable', 'passive', 'exclusive', 'auto_delete'].each do |opt|
            @config["queue_#{opt}"] = to_boolean(@config["queue_#{opt}"])
          end

          # Convert the queue parameters into AMQP-friendly sym-keyed
          # Hash that can be passed directly to underlying AMQP gem
          # methods
          @queue_config = @config.each_with_object({}) do |(key, value), result|
            md = /queue_(.*)/.match(key.to_s)
            result[md[1].to_sym] = value unless md.nil?
          end
        end

        # RFlow-called method at startup.
        # @return [void]
        def run!
          ::AMQP.connect(:host => @config['server'], :port => @config['port'], :vhost => @config['vhost'],
                         :username => @config['username'], :password => @config['password']) do |conn|
            @amqp_connection = conn
            RFlow.logger.info 'Connected to AMQP server...'

            conn.on_disconnection do
              RFlow.logger.error 'AMQP disconnected. Reconnecting...'
              log_port.send_message(RFlow::Message.new('RFlow::Message::Data::Log').tap do |m|
                m.data.timestamp = Integer(Time.now.to_f * 1000) # ms since epoch
                m.data.level = 'INFO'
                m.data.text = "AMQP disconnected. Reconnecting in #{@config['reconnect_interval']} seconds."
              end)

              EventMachine::Timer.new(@config['reconnect_interval']) { run! }
            end

            ::AMQP::Channel.new(@amqp_connection) do |channel|
              @amqp_channel = channel
              channel.auto_recovery = true
              @amqp_exchange = @amqp_channel.topic

              ::AMQP::Queue.new(@amqp_channel, @config['queue_name'], @queue_config) do |queue|
                @amqp_queue = queue
                @amqp_queue.bind(@amqp_exchange, :routing_key => @config['binding_pattern']).subscribe(:ack => true) do |header, payload|
                  RFlow.logger.debug { "#{name}: AMQP message received" }
                  processing_event = RFlow::Message::ProcessingEvent.new(uuid, Time.now.utc).tap do |e|
                    e.completed_at = Time.now.utc
                  end

                  amqp_port.send_message(RFlow::Message.new('RFlow::Message::Data::AMQP::Message').tap do |m|
                    header.to_hash.each {|k,v| m.data.header[k.to_s] = v }
                    m.data.payload = payload
                    m.provenance << processing_event
                  end)

                  # TODO: Optimize out if not connected
                  raw_port.send_message(RFlow::Message.new('RFlow::Message::Data::Raw').tap do |m|
                    m.data.raw = payload
                    m.provenance << processing_event
                  end)

                  header.ack
                end
              end
            end
          end
        end

        # @!visibility private
        def to_boolean(string)
          case string
          when /^true$/i, '1', true; true
          when /^false/i, '0', false; false
          else raise ArgumentError, "'#{string}' cannot be coerced to a boolean value"
          end
        end
      end
    end
  end
end
